from . import BaseActor
from ltr.models.loss.target_regression import DenseIouPred
import torch
from ltr.data.processing_utils import iou_xyxy
import cv2

class TransEncoderActor(BaseActor):
    """Actor for jointly training cls_18, cls_72 branch and regression branch, except for regression optimizer"""
    def __init__(self, net, objective, device, output_sz, loss_weight=None):
        super().__init__(net, objective, device)
        self.loss_weight = loss_weight
        self.output_sz = output_sz

    def box_cxcywh_to_xyxy(self, x):
        x_c, y_c, w, h = x.unbind(-1)
        b = [(x_c - 0.5 * w), (y_c - 0.5 * h),
             (x_c + 0.5 * w), (y_c + 0.5 * h)]
        return torch.stack(b, dim=-1)

    def box_xywh_to_xyxy(self, x):
        x1, y1, w, h = x.unbind(-1)
        b = [x1, y1, x1 + w, y1 + h]
        return torch.stack(b, dim=-1)

    def __call__(self, data, vis=False):
        """
        args:
            data - The input data.
            gen_iou_map -
        returns:
            loss    - the training loss
            stats  -  dict containing detailed losses
        """
        # Run network
        output_coord = self.net(train_imgs=data['train_images'],
                               test_imgs=data['test_images'],
                               train_bb=data['train_anno'])

        ### Regression offline loss ###
        # print(data['test_anno'].shape)
        # print(output_coord.shape)
        gt = self.box_xywh_to_xyxy(data['test_anno'].view(-1, 4)) / self.output_sz # X1Y1WH --> x1y1x2y2
        gt = gt.unsqueeze(-1)
        # output = output_coord.view(-1, 4)
        output = self.box_cxcywh_to_xyxy(output_coord.view(10, -1, 4)).permute(1, 2, 0) # XcYcWH --> x1y1x2y2  (nb, 4, 10)

        output_ious = iou_xyxy(gt, output) # nb, num_proposals
        max_ious, max_inds = torch.max(output_ious, dim=1, keepdim=True)
        # print("max_ious: {}, max_inds: {}, ious:{}".format(max_ious[0], max_inds[0], output_ious[0]))
        max_inds = max_inds.unsqueeze(1).repeat(1, 4, 1)  # nb, 4, 1
        output_max = torch.gather(output, dim=2, index=max_inds)  # nb, 4, 1

        # draw_img = None
        vis_results = {}
        if vis:
            img = self.convert_img_tensor_to_vis_format(data['test_images'])
            output_bb = output[0].clone() * self.output_sz
            bbs = output_bb.permute(1, 0).cpu().int().numpy() # 10, 4
            draw_img = self.draw_image(img, bbs)  # 1, c, h, w
            vis_results['vis_img'] = draw_img

        l1_loss = self.objective['l1'](output_max, gt) * self.loss_weight['l1']

        giou_loss, iou = self.objective['giou'](output_max, gt)
        iou = iou.mean()
        giou_loss = giou_loss * self.loss_weight['giou']

        # Total loss
        loss = l1_loss + giou_loss


        # Log stats
        stats = {'Loss/total': loss.item(),
                 'Loss/l1': l1_loss.item(),
                 'Loss/giou': giou_loss.item(),
                 'mean_iou': iou.item()
                }


        return loss, stats, vis_results

    def convert_img_tensor_to_vis_format(self, img_tensors):
        """
        :param img_tensors: n, b, c, h, w
        :return:
        """
        _mean = torch.Tensor([0.485, 0.456, 0.406]).view(-1, 1, 1)
        _std = torch.Tensor([0.229, 0.224, 0.225]).view(-1, 1, 1)
        img_tensor = img_tensors[0, 0].clone().cpu() * _std + _mean
        img_tensor *= 255.0
        img = img_tensor.permute(1, 2, 0).numpy()  # h, w, 3
        return img

    def draw_image(self, img, bboxs):
        """Display a 2D tensor.
        """
        img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
        colors = [(0,0,255),
                  (0,255,0),
                  (255,0,0),
                  (255,255,0),
                  (255,0,255),
                  (0,255,255),
                  (0,0,125),
                  (0,125,0),
                  (125,0,0),
                  (125,125,125)]
        i = 0
        for bbox in bboxs:
            img = cv2.rectangle(img, (bbox[0], bbox[1]), (bbox[2], bbox[3]), colors[i], 2)
            i += 1
        # draw_rec = cv2.circle(draw_rec, (center[0], center[1]), 3, (0, 0, 255), -1)
        # draw_rec = cv2.circle(draw_rec, (center2[1], center2[0]), 3, (0, 255, 0), -1)

        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img_tensor = torch.from_numpy(img).float().permute(2, 0, 1).unsqueeze(0)
        return img_tensor

