import math
import torch.nn as nn
from collections import OrderedDict
import ltr.models.target_classifier.linear_filter as target_clf
import ltr.models.target_classifier.features as clf_features
import ltr.models.target_classifier.initializer as clf_initializer
import ltr.models.target_classifier.optimizer as clf_optimizer
import ltr.models.target_regression.reg_optimizer as reg_optimizer
import ltr.models.backbone as backbones
from ltr.models.target_regression.reg_online_filter import RegFilter
from ltr.models.target_classifier.up_features import FPNUpBlock
from ltr import model_constructor
import torch.nn.functional as F
from ltr.models.layers.non_local import NONLocalBlock2D
# from ltr.models.target_classifier.nonlocal_classifier import NonlocalClassifier
from ltr.models.target_regression.nonlocal_regressor import NonlocalRegressor
import torch
import os, cv2
import torch.distributed as dist
from ltr.models.TransEncoder.transformer import Transformer
from ltr.models.target_classifier.initializer import FilterPool
from ltr.models.TransEncoder.head import build_box_head

class TransClsNet(nn.Module):
    """The FCOTNet network.
    args:
        feature_extractor:  Backbone feature extractor network. Must return a dict of feature maps.
        classification_layer:  Name of the backbone feature layer to use for classification.
        pyramid_*: The blocks in Decoder.
        classifier_18:  Target classification module with feature size of 18.
        classifier_72:  Target classification module with feature size of 72.
        regressor_72:  Anchor-free regression module with feature size of 72.
        train_cls_72_and_reg_init: Whether clssifier_72 and regression initial model should be trained or not.
        train_reg_optimizer:  Whether regression optimizer should be trained or not.
        train_cls_18:  Whether classifier_18 should be trained or not. """

    def __init__(self, feature_extractor, classification_layer, classifier_s, hidden_dim=256):
        super().__init__()

        self.feature_extractor = feature_extractor
        self.transformer = classifier_s

        # self.proj1 = nn.Linear(256, hidden_dim)
        # self.proj2 = nn.Linear(512, hidden_dim)
        self.proj3 = nn.Linear(1024, hidden_dim)
        self.target_pool = FilterPool(filter_size=4, feature_stride=16, pool_square=False)
        self.score_head = build_box_head(hidden_dim, outdim=1)

        self.classification_layer = [classification_layer] if isinstance(classification_layer, str) else classification_layer
        self.output_layers = ['layer1', 'layer2', 'layer3']

        # for func_name in ['classifier_18', 'feature_extractor']:
        #     for p in getattr(self, func_name).parameters():
        #         p.requires_grad_(False)
        # for func_name in ['classifier_72', 'regressor_72', 'pyramid_first_conv', 'pyramid_36', 'pyramid_72']:
        #     for p in getattr(self, func_name).parameters():
        #         p.requires_grad_(False)

    def forward(self, train_imgs, test_imgs, train_bb, *args, **kwargs):

        assert train_imgs.dim() == 5 and test_imgs.dim() == 5, 'Expect 5 dimensional inputs'
        n, b, _, _, _ = test_imgs.size()

        # Extract backbone features
        train_feat = self.extract_backbone_features(train_imgs.view(-1, *train_imgs.shape[-3:]),
                                                    layers=['layer1','layer2','layer3'])
        test_feat = self.extract_backbone_features(test_imgs.view(-1, *test_imgs.shape[-3:]),
                                                   layers=['layer1','layer2','layer3'])

        # Classification features
        train_feat_s = self.proj3(train_feat['layer3'].permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
        test_feat_s = self.proj3(test_feat['layer3'].permute(0, 2, 3, 1)).permute(0, 3, 1, 2)
        # train_feat_s = train_feat['layer3']
        # test_feat_s = test_feat['layer3']

        # generate score_map
        # target0 = train_feat_s.view(n, b, *train_feat_s.shape[-3:])[0]
        target = train_feat_s
        target = self.generate_target(target, train_bb)  # NB, C, Ht, Wt
        # target = self.generate_target(target0, train_bb[0])  # NB, C, Ht, Wt
        # target = target.unsqueeze(0).repeat(n, 1, 1, 1, 1).view(n*b, -1, 4, 4)
        target = target.flatten(2).permute(2, 0, 1)
        search = test_feat_s.flatten(start_dim=2).permute(2, 0, 1)  # HWxNBxC
        hs = self.transformer(search, target)  # HWxNBxC
        score_s = self.score_head(hs).squeeze().permute(1, 0).view(n, b, *test_feat_s.shape[-2:])

        output = {}
        output['scores_s'] = score_s

        return output

    def generate_target(self, feat_train, bb):
        bb_pool = bb.clone()
        # print("bb: {}".format(bb))
        target = self.target_pool(feat_train, bb_pool)

        # target = target.view(-1, self.filter_dim, target.shape[-3] // self.filter_dim, target.shape[-2], target.shape[-1])
        return target


    def get_backbone_clf_feat(self, backbone_feat):
        feat = OrderedDict({l: backbone_feat[l] for l in self.classification_layer})
        if len(self.classification_layer) == 1:
            return feat[self.classification_layer[0]]
        return feat

    def extract_backbone_features(self, im, layers=None):
        if layers is None:
            layers = self.output_layers
        return self.feature_extractor(im, layers)

    def extract_features(self, im, layers=None):
        if layers is None:
            layers = self.bb_regressor_layer + ['classification']
        if 'classification' not in layers:
            return self.feature_extractor(im, layers)
        backbone_layers = sorted(list(set([l for l in layers + self.classification_layer if l != 'classification'])))
        all_feat = self.feature_extractor(im, backbone_layers)
        all_feat['classification'] = self.extract_classification_feat(all_feat)
        return OrderedDict({l: all_feat[l] for l in layers})

def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True

def get_rank():
    if not is_dist_avail_and_initialized():
        return 0
    return dist.get_rank()

def is_main_process():
    return get_rank() == 0

@model_constructor
def transClsnet(clf_filter_size=4, optim_iter=5, optim_init_step=1.0, optim_init_reg=0.01,
            classification_layer='layer3', feat_stride=16, clf_feat_blocks=0,
            clf_feat_norm=True, init_filter_norm=False, final_conv=True, out_feature_dim=512,
            init_gauss_sigma=1.0, num_dist_bins=5, bin_displacement=1.0, mask_init_factor=4.0,
            score_act='relu', act_param=None, target_mask_act='sigmoid', detach_length=float('Inf'),
            frozen_backbone_layers=(), freeze_bn=True):
    # Backbone
    backbone_net = backbones.resnet50(pretrained=is_main_process(), frozen_layers=frozen_backbone_layers,
                                      freeze_bn=freeze_bn)
    # static classifier (layer1-18)
    transformer = Transformer(d_model=256, nhead=8, num_encoder_layers=6, num_decoder_layers=6)

    net = TransClsNet(feature_extractor=backbone_net, classification_layer=classification_layer,
                  classifier_s=transformer, hidden_dim=256)
    return net