import torch.optim as optim
import torchvision.transforms
import torch
from ltr.dataset import Lasot, Got10k, TrackingNet, MSCOCOSeq
from ltr.data import sampler, LTRLoader, processing_fcot
from ltr.models.tracking import transEncoder
import ltr.models.loss as ltr_losses
from ltr.models.loss.target_regression import REGLoss
from ltr import actors
from ltr.models.loss.giou import giou_loss
from torch.nn.functional import l1_loss
from ltr.trainers import LTRTrainer
import ltr.data.transforms as dltransforms
from ltr import MultiGPU
import ltr.data.new_transform as tfm
import collections
from torch.utils.data.distributed import DistributedSampler
from torch.nn.parallel import DistributedDataParallel as DDP
import time

# Address the issue: "RuntimeError: received 0 items of ancdata"
import resource
rlimit = resource.getrlimit(resource.RLIMIT_NOFILE)
resource.setrlimit(resource.RLIMIT_NOFILE, (2048, rlimit[1]))
torch.multiprocessing.set_sharing_strategy('file_system')


def run(settings):
    settings.description = 'Default train settings for Transformer-Encoder.'
    # settings.use_pretrained_super_dimp = True
    # settings.pretrained_super_dimp50 = "../models/super_dimp.pth.tar"
    settings.multi_gpu = True
    settings.print_interval = 10
    settings.normalize_mean = [0.485, 0.456, 0.406]
    settings.normalize_std = [0.229, 0.224, 0.225]
    settings.search_area_factor = 5.0
    settings.output_sigma_factor = 1/4
    settings.clf_target_filter_sz = 4
    settings.reg_target_filter_sz = 3
    settings.feature_sz = 18
    settings.output_sz = settings.feature_sz * 16
    settings.center_jitter_factor = {'train': 3, 'test': 4.5}
    settings.scale_jitter_factor = {'train': 0.25, 'test': 0.5}
    settings.hinge_threshold = 0.05
    settings.logging_file = 'log.txt'

    # Train datasets
    lasot_train = Lasot(settings.env.lasot_dir, split='train')
    got10k_train = Got10k(settings.env.got10k_dir, split='vottrain')
    trackingnet_train = TrackingNet(settings.env.trackingnet_dir, set_ids=list(range(4)))
    coco_train = MSCOCOSeq(settings.env.coco_dir)

    # Validation datasets
    got10k_val = Got10k(settings.env.got10k_dir, split='votval')

    # Data transform
    transform_joint = tfm.Transform(tfm.ToGrayscale(probability=0.05),
                                    tfm.RandomHorizontalFlip(probability=0.5))

    transform_train = tfm.Transform(tfm.ToTensorAndJitter(0.2),
                                    tfm.RandomHorizontalFlip(probability=0.5),
                                    tfm.Normalize(mean=settings.normalize_mean, std=settings.normalize_std))

    transform_val = tfm.Transform(tfm.ToTensor(),
                                  tfm.Normalize(mean=settings.normalize_mean, std=settings.normalize_std))

    # The tracking pairs processing module
    output_sigma = settings.output_sigma_factor / settings.search_area_factor
    label_params = {'feature_sz': settings.feature_sz, 'sigma_factor': output_sigma, 'kernel_sz': settings.clf_target_filter_sz}
    data_processing_train = processing_fcot.TransEncoderProcessing(search_area_factor=settings.search_area_factor,
                                                                      output_sz=settings.output_sz,
                                                                      output_w=88,
                                                                      output_h=88,
                                                                      center_jitter_factor=settings.center_jitter_factor,
                                                                      scale_jitter_factor=settings.scale_jitter_factor,
                                                                      mode='sequence',
                                                                      max_scale_change=1.5,
                                                                      crop_type='inside_major',
                                                                      output_spatial_scale=1 / 4.,
                                                                      label_function_params=label_params,
                                                                      transform=transform_train,
                                                                      joint_transform=transform_joint)

    data_processing_val = processing_fcot.TransEncoderProcessing(search_area_factor=settings.search_area_factor,
                                                                    output_sz=settings.output_sz,
                                                                    output_w=88,
                                                                    output_h=88,
                                                                    center_jitter_factor=settings.center_jitter_factor,
                                                                    scale_jitter_factor=settings.scale_jitter_factor,
                                                                    mode='sequence',
                                                                    crop_type='inside_major',
                                                                    max_scale_change=1.5,
                                                                    output_spatial_scale=1 / 4.,
                                                                    label_function_params=label_params,
                                                                    transform=transform_val,
                                                                    joint_transform=transform_joint)

    # Train sampler and loader
    dataset_train = sampler.FCOTSampler([lasot_train, got10k_train, trackingnet_train, coco_train], [settings.lasot_rate,1,1,1],
                                        samples_per_epoch=settings.samples_per_epoch * settings.batch_size, max_gap=100, num_test_frames=1, num_train_frames=1,
                                        processing=data_processing_train)
    # dataset_train = sampler.FCOTSampler([got10k_train],
    #                                     [1],
    #                                     samples_per_epoch=settings.samples_per_epoch, max_gap=200, num_test_frames=1,
    #                                     num_train_frames=1,
    #                                     processing=data_processing_train)
    train_sampler = DistributedSampler(dataset_train) if settings.local_rank != -1 else None
    shuffle = False if settings.local_rank != -1 else True
    loader_train = LTRLoader('train', dataset_train, training=True, batch_size=settings.batch_size, sampler=train_sampler,
                             num_workers=settings.num_workers, shuffle=shuffle, drop_last=True, stack_dim=1)


    # Validation samplers and loaders
    dataset_val = sampler.FCOTSampler([got10k_val], [1], samples_per_epoch=20000, max_gap=100, num_test_frames=1,
                                      num_train_frames=1, processing=data_processing_val)
    val_sampler = DistributedSampler(dataset_val) if settings.local_rank != -1 else None
    loader_val = LTRLoader('val', dataset_val, training=False, batch_size=settings.batch_size, epoch_interval=5,
                           sampler=val_sampler, num_workers=settings.num_workers, shuffle=False, drop_last=True, stack_dim=1)

    # Create network
    net = transEncoder.transEncoder(classification_layer='layer3',
                                    freeze_bn=True,
                                    backbone_pretrained=True,
                                    frozen_backbone_layers=['conv1', 'bn1', 'layer1', 'layer2'])

    # Load dimp-model as initial weights
    # device = torch.device('cuda:{}'.format(settings.devices_id[0]) if torch.cuda.is_available() else 'cpu')
    # if settings.use_pretrained_super_dimp:
    #     assert settings.pretrained_super_dimp50 is not None
    #     super_dimp50 = torch.load(settings.pretrained_super_dimp50, map_location=device)
    #     state_dict = collections.OrderedDict()
    #     for key, v in super_dimp50['net'].items():
    #         if key.split('.')[0] == 'feature_extractor':
    #             state_dict['.'.join(key.split('.')[1:])] = v
    #
    #     net.feature_extractor.load_state_dict(state_dict)
    #
    #     state_dict = collections.OrderedDict()
    #     for key, v in super_dimp50['net'].items():
    #         if key.split('.')[0] == 'classifier':
    #             state_dict['.'.join(key.split('.')[1:])] = v
    #     net.classifier_18.load_state_dict(state_dict)
    #     print("loading backbone and Classifier modules from super-DiMP50 done.")

    # # Wrap the network for multi GPU training
    # if settings.multi_gpu:
    #     net = MultiGPU(net, device_ids=settings.devices_id, dim=1).to(device)

    net.cuda()
    if settings.local_rank != -1:
        net = torch.nn.SyncBatchNorm.convert_sync_batchnorm(net)
        net = DDP(net, device_ids=[settings.local_rank], find_unused_parameters=True)
        settings.device = torch.device("cuda:%d" % settings.local_rank)
    else:
        settings.device = torch.device("cuda:0")

    # Loss for cls_72, cls_18 and regression
    objective = {'giou': giou_loss, 'l1': l1_loss}

    # Create actor and adam-optimizer
    loss_weight = {'giou': 2.0, 'l1': 5.0}
    actor = actors.TransEncoderActor(net=net, objective=objective, output_sz=settings.output_sz, loss_weight=loss_weight, device=settings.device)

    # print(actor.net)
    # optimizer = optim.AdamW([{'params': actor.net.module.feature_extractor.parameters(), 'lr': 5e-5},
    #                          {'params': actor.net.module.proj.parameters()},
    #                          {'params': actor.net.module.transformer.parameters()},
    #                          {'params': actor.net.module.box_head.parameters()}], lr=0.0001, weight_decay=0.0001)
    optimizer = optim.AdamW([{'params': actor.net.parameters()}], lr=0.0001, weight_decay=0.0001)

    lr_scheduler = optim.lr_scheduler.MultiStepLR(optimizer, milestones=[200, 250], gamma=0.2)

    trainer = LTRTrainer(actor, [loader_train, loader_val], optimizer, settings, settings.device, lr_scheduler,
                             logging_file=settings.logging_file, max_norm=0.1)

    trainer.train(settings.total_epochs, load_latest=True, fail_safe=True)