import torch.optim as optim
import torchvision.transforms
import torch
import torch.nn as nn
from ltr.dataset import LasotMask, Got10kMask, TrackingNet, MSCOCOSeqMask, Saliency, Youtube_VOS
from ltr.data import mask_sampler, LTRLoader, processing_mask
from ltr.models.tracking import trSeg2
import ltr.models.loss as ltr_losses
from ltr.models.loss.target_regression import REGLoss
from ltr import actors
from ltr.models.loss.giou import giou_loss
from torch.nn.functional import l1_loss
from ltr.trainers import LTRMaskTrainer
import ltr.data.transform_mask as dltransforms
from ltr import MultiGPU
import ltr.data.new_transform as tfm
import collections
from torch.utils.data.distributed import DistributedSampler
from torch.nn.parallel import DistributedDataParallel as DDP
import time

# Address the issue: "RuntimeError: received 0 items of ancdata"
import resource
rlimit = resource.getrlimit(resource.RLIMIT_NOFILE)
resource.setrlimit(resource.RLIMIT_NOFILE, (2048, rlimit[1]))
torch.multiprocessing.set_sharing_strategy('file_system')


def run(settings):
    settings.description = 'Default train settings for Transformer-seg-tracker.'
    settings.multi_gpu = True
    settings.print_interval = 1
    settings.normalize_mean = [0.485, 0.456, 0.406]
    settings.normalize_std = [0.229, 0.224, 0.225]
    settings.search_area_factor = 2.0
    settings.output_sigma_factor = 1/4
    settings.feature_sz = 16
    settings.output_sz = settings.feature_sz * 16
    settings.center_jitter_factor = {'train': 0.25, 'test': 0.25}
    settings.scale_jitter_factor = {'train': 0.25, 'test': 0.25}
    settings.hinge_threshold = 0.05
    settings.logging_file = 'log.txt'

    # Train datasets
    lasot_train = LasotMask(settings.env.lasot_dir, split='train')
    got10k_train = Got10kMask(settings.env.got10k_dir, split='vottrain')
    # trackingnet_train = TrackingNet(settings.env.trackingnet_dir, set_ids=list(range(4)))
    coco_train = MSCOCOSeqMask(settings.env.coco_dir)

    # imagenet_vid = ImagenetVID()
    # imagenet_det = ImagenetDET()

    # - mask datasets
    youtube_vos = Youtube_VOS(settings.env.youtube_vos_dir)
    saliency = Saliency(settings.env.saliency_dir)

    # Validation datasets
    # got10k_val = Got10k(settings.env.got10k_dir, split='votval')

    # Data transform
    # The joint augmentation transform, that is applied to the pairs jointly
    transform_joint = dltransforms.ToGrayscale(probability=0.05)
    # The augmentation transform applied to the training set (individually to each image in the pair)
    transform_train = torchvision.transforms.Compose([dltransforms.ToTensorAndJitter(0.2),
                                                      torchvision.transforms.Normalize(mean=settings.normalize_mean,
                                                                                       std=settings.normalize_std)])
    transform_val = torchvision.transforms.Compose([torchvision.transforms.ToTensor(),
                                                    torchvision.transforms.Normalize(mean=settings.normalize_mean,
                                                                                     std=settings.normalize_std)])
    # transform_joint = tfm.Transform(tfm.ToGrayscale(probability=0.05),
    #                                 tfm.RandomHorizontalFlip(probability=0.5))
    #
    # transform_train = tfm.Transform(tfm.ToTensorAndJitter(0.2),
    #                                 tfm.RandomHorizontalFlip(probability=0.5),
    #                                 tfm.Normalize(mean=settings.normalize_mean, std=settings.normalize_std))
    #
    # transform_val = tfm.Transform(tfm.ToTensor(),
    #                               tfm.Normalize(mean=settings.normalize_mean, std=settings.normalize_std))

    # The tracking pairs processing module
    # output_sigma = settings.output_sigma_factor / settings.search_area_factor
    # label_params = {'feature_sz': settings.feature_sz, 'sigma_factor': output_sigma, 'kernel_sz': settings.clf_target_filter_sz}
    data_processing_train = processing_mask.SEMaskProcessing(search_area_factor=settings.search_area_factor,
                                                          output_sz=settings.output_sz,
                                                          center_jitter_factor=settings.center_jitter_factor,
                                                          scale_jitter_factor=settings.scale_jitter_factor,
                                                          mode='sequence',
                                                          transform=transform_train,
                                                          joint_transform=transform_joint)

    data_processing_val = processing_mask.SEMaskProcessing(search_area_factor=settings.search_area_factor,
                                                        output_sz=settings.output_sz,
                                                        center_jitter_factor=settings.center_jitter_factor,
                                                        scale_jitter_factor=settings.scale_jitter_factor,
                                                        mode='sequence',
                                                        transform=transform_val,
                                                        joint_transform=transform_joint)

    # Train sampler and loader
    dataset_train = mask_sampler.MaskSampler([youtube_vos, saliency],
                        [2, 3],
                        samples_per_epoch=settings.samples_per_epoch * settings.batch_size,
                        max_gap=50,
                        processing=data_processing_train)
    # dataset_train = mask_sampler.MaskSampler([saliency],
    #                                          [1],
    #                                          samples_per_epoch=settings.samples_per_epoch * settings.batch_size,
    #                                          max_gap=50,
    #                                          processing=data_processing_train)
    train_sampler = DistributedSampler(dataset_train) if settings.local_rank != -1 else None
    shuffle = False if settings.local_rank != -1 else True
    loader_train = LTRLoader('train', dataset_train, training=True, batch_size=settings.batch_size, sampler=train_sampler,
                             num_workers=settings.num_workers, shuffle=shuffle, drop_last=True, stack_dim=1)

    # Validation samplers and loaders
    lasot_test = LasotMask(split='test')
    dataset_val = mask_sampler.MaskSampler([lasot_test], [1], samples_per_epoch=500 * settings.batch_size, max_gap=50,
                                          processing=data_processing_val)
    val_sampler = DistributedSampler(dataset_val) if settings.local_rank != -1 else None
    loader_val = LTRLoader('val', dataset_val, training=False, batch_size=settings.batch_size, epoch_interval=5,
                           sampler=val_sampler, num_workers=settings.num_workers, shuffle=False, drop_last=True, stack_dim=1)

    # Create network
    net = trSeg2.transSeg(classification_layer='layer3', backbone_pretrained=True, freeze_bn=False,
                         frozen_backbone_layers=['conv1', 'bn1', 'layer1', 'layer2'])

    net.cuda()
    if settings.local_rank != -1:
        net = torch.nn.SyncBatchNorm.convert_sync_batchnorm(net)
        net = DDP(net, device_ids=[settings.local_rank], find_unused_parameters=True)
        settings.device = torch.device("cuda:%d" % settings.local_rank)
    else:
        settings.device = torch.device("cuda:0")

    # Loss for cls_72, cls_18 and regression
    objective = {'giou': giou_loss, 'l1': l1_loss, 'mask': nn.BCELoss()}

    # Create actor and adam-optimizer
    loss_weight = {'giou': 2.0, 'l1': 5.0, 'mask': 1.0}
    actor = actors.TrSegActor(net=net, objective=objective, output_sz=settings.output_sz, loss_weight=loss_weight, device=settings.device)

    # print(actor.net)
    # optimizer = optim.AdamW([{'params': actor.net.module.feature_extractor.parameters(), 'lr': 5e-5},
    #                          {'params': actor.net.module.proj.parameters()},
    #                          {'params': actor.net.module.transformer.parameters()},
    #                          {'params': actor.net.module.box_head.parameters()}], lr=0.0001, weight_decay=0.0001)
    optimizer = optim.AdamW([{'params': actor.net.parameters()}], lr=0.0001, weight_decay=0.0001)

    lr_scheduler = optim.lr_scheduler.MultiStepLR(optimizer, milestones=[138, 148], gamma=0.2)

    trainer = LTRMaskTrainer(actor, [loader_train, loader_val], optimizer, settings, settings.device, lr_scheduler,
                             logging_file=settings.logging_file, max_norm=0.1)

    trainer.train(settings.total_epochs, load_latest=True, fail_safe=True)