import os
import numpy as np
import json
import sys
import glob

env_path = os.path.join(os.path.dirname(__file__), '..')
if env_path not in sys.path:
    sys.path.append(env_path)

from pytracking.evaluation.otbdataset import OTBDataset
import argparse

def evaluate_otb(dataset, results_dir, report_file, delimiter='\t'):
    performance = {}
    seq_num = len(dataset)
    nbins_iou = 21
    nbins_ce = 51
    succ_curve = np.zeros((seq_num, nbins_iou))
    prec_curve = np.zeros((seq_num, nbins_ce))

    print('-----------Evaluating on OTB100------------')

    performance.update({
        'overall': {},
        'seq_wise': {}})

    for s, seq in enumerate(dataset):
        seq_name = seq.name
        anno = seq.ground_truth_rect[:, :]
        record_file = os.path.join(results_dir, '%s.txt' % seq_name)
        boxes = np.loadtxt(record_file, delimiter=delimiter)
        boxes[0] = anno[0]
        assert len(boxes) == len(anno)

        ious, center_errors = _calc_metrics(boxes, anno)
        succ_curve[s], prec_curve[s] = _calc_curves(ious, center_errors, nbins_iou, nbins_ce)

        # store sequence-wise performance
        performance['seq_wise'].update({seq_name: {
            'success_score': np.mean(succ_curve[s]),
            'precision_score': prec_curve[s][20]}})

    succ_curve = np.mean(succ_curve, axis=0)
    prec_curve = np.mean(prec_curve, axis=0)
    succ_score = np.mean(succ_curve)
    prec_score = prec_curve[20]

    # store overall performance
    performance['overall'].update({
        'success_score': succ_score,
        'precision_score': prec_score})

    # report the performance
    with open(report_file, 'w') as f:
        json.dump(performance, f, indent=4)

    return succ_score, prec_score


def _calc_metrics(boxes, anno):
    ious = rect_iou(boxes, anno)
    center_errors = center_error(boxes, anno)
    return ious, center_errors


def center_error(rects1, rects2):
    r"""Center error.
    Args:
        rects1 (numpy.ndarray): An N x 4 numpy array, each line represent a rectangle
            (left, top, width, height).
        rects2 (numpy.ndarray): An N x 4 numpy array, each line represent a rectangle
            (left, top, width, height).
    """
    centers1 = rects1[..., :2] + (rects1[..., 2:] - 1) / 2
    centers2 = rects2[..., :2] + (rects2[..., 2:] - 1) / 2
    errors = np.sqrt(np.sum(np.power(centers1 - centers2, 2), axis=-1))

    return errors


def rect_iou(rects1, rects2, bound=None):
    r"""Intersection over union.
    Args:
        rects1 (numpy.ndarray): An N x 4 numpy array, each line represent a rectangle
            (left, top, width, height).
        rects2 (numpy.ndarray): An N x 4 numpy array, each line represent a rectangle
            (left, top, width, height).
        bound (numpy.ndarray): A 4 dimensional array, denotes the bound
            (min_left, min_top, max_width, max_height) for ``rects1`` and ``rects2``.
    """
    assert rects1.shape == rects2.shape
    if bound is not None:
        # bounded rects1
        rects1[:, 0] = np.clip(rects1[:, 0], 0, bound[0])
        rects1[:, 1] = np.clip(rects1[:, 1], 0, bound[1])
        rects1[:, 2] = np.clip(rects1[:, 2], 0, bound[0] - rects1[:, 0])
        rects1[:, 3] = np.clip(rects1[:, 3], 0, bound[1] - rects1[:, 1])
        # bounded rects2
        rects2[:, 0] = np.clip(rects2[:, 0], 0, bound[0])
        rects2[:, 1] = np.clip(rects2[:, 1], 0, bound[1])
        rects2[:, 2] = np.clip(rects2[:, 2], 0, bound[0] - rects2[:, 0])
        rects2[:, 3] = np.clip(rects2[:, 3], 0, bound[1] - rects2[:, 1])

    rects_inter = _intersection(rects1, rects2)
    areas_inter = np.prod(rects_inter[..., 2:], axis=-1)

    areas1 = np.prod(rects1[..., 2:], axis=-1)
    areas2 = np.prod(rects2[..., 2:], axis=-1)
    areas_union = areas1 + areas2 - areas_inter

    eps = np.finfo(float).eps
    ious = areas_inter / (areas_union + eps)
    ious = np.clip(ious, 0.0, 1.0)

    return ious


def _intersection(rects1, rects2):
    r"""Rectangle intersection.
    Args:
        rects1 (numpy.ndarray): An N x 4 numpy array, each line represent a rectangle
            (left, top, width, height).
        rects2 (numpy.ndarray): An N x 4 numpy array, each line represent a rectangle
            (left, top, width, height).
    """
    assert rects1.shape == rects2.shape
    x1 = np.maximum(rects1[..., 0], rects2[..., 0])
    y1 = np.maximum(rects1[..., 1], rects2[..., 1])
    x2 = np.minimum(rects1[..., 0] + rects1[..., 2],
                    rects2[..., 0] + rects2[..., 2])
    y2 = np.minimum(rects1[..., 1] + rects1[..., 3],
                    rects2[..., 1] + rects2[..., 3])

    w = np.maximum(x2 - x1, 0)
    h = np.maximum(y2 - y1, 0)

    return np.stack([x1, y1, w, h]).T


def _calc_curves(ious, center_errors, nbins_iou, nbins_ce):
    ious = np.asarray(ious, float)[:, np.newaxis]
    center_errors = np.asarray(center_errors, float)[:, np.newaxis]

    thr_iou = np.linspace(0, 1, nbins_iou)[np.newaxis, :]
    thr_ce = np.arange(0, nbins_ce)[np.newaxis, :]

    bin_iou = np.greater(ious, thr_iou)
    bin_ce = np.less_equal(center_errors, thr_ce)

    succ_curve = np.mean(bin_iou, axis=0)
    prec_curve = np.mean(bin_ce, axis=0)

    return succ_curve, prec_curve

def main():
    parser = argparse.ArgumentParser(description='Evalute tracking results of OTB100 dataset.')
    parser.add_argument('otb_prefix', type=str, help='Path of otb results generated by FCOT.')
    parser.add_argument('--delimiter', type=str, default='\t', help='Type for saving tracking results.')
    parser.add_argument('--base_results_path', type=str, default='/home/cyt/data/FCOT_results/otb_results/fcot')
    args = parser.parse_args()

    otb_dataset = OTBDataset()

    args.otb_prefix = args.otb_prefix + '*'
    for otb_results_path in glob.glob(os.path.join(args.base_results_path, args.otb_prefix)):
        print(otb_results_path)
        succ_score, prec_score = evaluate_otb(otb_dataset,
                                              results_dir=otb_results_path,
                                              report_file=os.path.join(otb_results_path, 'otb100-performance.txt'),
                                              delimiter=args.delimiter)
        print("succ_score-{}, prec_score-{}".format(succ_score, prec_score))
    print("Evaluate tracker on OTB100 done!")


if __name__ == '__main__':
    main()
