import numpy as np
import time
import torch
import torch.nn.functional as F
import numpy as np
import cv2
from collections import deque

from pytracking.utils.loading import load_network
from ltr.data.processing_utils_mask import sample_target_SE, transform_image_to_crop_SE, map_mask_back

def rect_from_mask(mask):
    """
    mask: (H, W)
    create an axis-aligned rectangle from a given binary mask
    mask in created as a minimal rectangle containing all non-zero pixels
    """
    x_ = np.sum(mask, axis=0)
    y_ = np.sum(mask, axis=1)
    x0 = np.min(np.nonzero(x_))
    x1 = np.max(np.nonzero(x_))
    y0 = np.min(np.nonzero(y_))
    y1 = np.max(np.nonzero(y_))
    return [x0, y0, x1 - x0 + 1, y1 - y0 + 1]


def mask_from_rect(rect, output_sz):
    """
    create a binary mask from a given rectangle
    rect: axis-aligned rectangle [x0, y0, width, height]
    output_sz: size of the output [width, height]
    """
    mask = np.zeros((output_sz[1], output_sz[0]), dtype=np.uint8)
    x0 = max(int(round(rect[0])), 0)
    y0 = max(int(round(rect[1])), 0)
    x1 = min(int(round(rect[0] + rect[2])), output_sz[0])
    y1 = min(int(round(rect[1] + rect[3])), output_sz[1])
    mask[y0:y1, x0:x1] = 1
    return mask

class TrSegModule(object):
    def __init__(self, trseg_net_dir, search_factor=2.0, input_sz=256):
        self.trseg_network = self.get_network(trseg_net_dir)
        self.search_factor = search_factor
        self.input_sz = input_sz
        self.mean = np.array([0.485, 0.456, 0.406]).reshape((1, 1, 3))
        self.std = np.array([0.229, 0.224, 0.225]).reshape((1, 1, 3))
        self.mask_targets = None

        self.online_mask_targets = None
        self.online_mem = deque()
        self.mask_mem_capacity = 550

    def initialize(self, frame1, bbox1):
        """
        Args:
            frame1(np.array): cv2 iamge array with shape (H,W,3)
            bbox1(np.array): with shape(4,)
        """

        """Step1: get cropped patch(tensor)"""
        patch1, h_f, w_f = sample_target_SE(frame1, bbox1, self.search_factor, self.input_sz,
                                            mode=cv2.BORDER_CONSTANT)
        patch1_tensor = self.img_preprocess(patch1)

        """Step2: get GT's cooridinate on the cropped patch(tensor)"""
        crop_sz = torch.Tensor((self.input_sz, self.input_sz))
        bbox1_tensor = self.gt_preprocess(bbox1)  # (4,)
        bbox1_crop_tensor = transform_image_to_crop_SE(bbox1_tensor, bbox1_tensor, h_f, w_f, crop_sz).cuda()

        """Step3: forward prop (targets branch)"""
        self.mask_targets = self.generate_targets(patch1_tensor, bbox1_crop_tensor)

        # self.static_mask_targets = self.mask_targets.clone() # 8*8, 1, 256
        # print('static mem shape: ', self.static_mask_targets.shape)
        
    
    def update_mask_targets(self, feats):
        # feats: n, 1, c
        feats_len = feats.shape[0]
        while feats_len > self.mask_mem_capacity:
            head_feat = self.online_mem.popleft()
            self.mask_mem_capacity += head_feat.shape[0]
        self.mask_mem_capacity -= feats_len
        self.online_mem.append(feats)

        self.online_mask_targets = torch.cat(list(self.online_mem), dim=0)
        

    def generate_targets(self, crop_patch, bbox):
        with torch.no_grad():
            targets_feat = self.trseg_network.feature_extractor(crop_patch.view(-1, *crop_patch.shape[-3:]),
                                                                ['conv1', 'layer1', 'layer2', 'layer3'])
            targets_feat = self.trseg_network.get_backbone_feat(targets_feat)  # n, c, h, w
            targets = self.trseg_network.generate_target(targets_feat, bbox)
            targets =  self.trseg_network.proj(targets.flatten(2).permute(2, 0, 1))  # hw, n, c

        return targets

    def generate_test_mask(self, test_patch, update_mask_target=False):

        with torch.no_grad():
            test_feat = self.trseg_network.feature_extractor(test_patch.view(-1, *test_patch.shape[-3:]),
                                                                ['conv1', 'layer1', 'layer2', 'layer3'])
            Lfeat_list = [feat for name, feat in test_feat.items() if name != 'layer3']
            test_feat = self.trseg_network.get_backbone_feat(test_feat)  # n, c, h, w
            n, _, hs, ws = test_feat.size()

            search = self.trseg_network.proj(test_feat.flatten(start_dim=2).permute(2, 0, 1))  # HsWs, N, C
            queried_sear = self.trseg_network.transformer(search, self.mask_targets)[0]  # HW, n, C  (C=64)
            queried_sear = queried_sear.permute(1, 2, 0).view(n, -1, hs, ws)  # n, 64, 16, 16
            outputs_corner = self.trseg_network.corner_head(queried_sear)  #(n, 4)-->[x0, y0, x1, y1]

            if self.online_mask_targets is not None:
                # online_weight = 0.2
                online_queried_sear = self.trseg_network.transformer(search, self.online_mask_targets)[0]
                online_queried_sear = online_queried_sear.permute(1, 2, 0).view(n, -1, hs, ws)
                online_outputs_mask = self.trseg_network.mask_head(online_queried_sear, Lfeat_list)
                # queried_sear = self.online_weight * online_queried_sear + (1-self.online_weight) * queried_sear

            outputs_mask = self.trseg_network.mask_head(queried_sear, Lfeat_list)  # (n, 1, 256, 256)

            if self.online_mask_targets is not None:
                outputs_mask = self.online_weight * online_outputs_mask + (1-self.online_weight) * outputs_mask

            # update_mask_target = False
            update_thre = 0.4
            if update_mask_target:
                self.online_weight = 0.2
                fore_feat_mask = F.interpolate(outputs_mask,size=(16,16),mode='bilinear').reshape(-1) # 256
                fore_feat_index = (fore_feat_mask > update_thre).nonzero().squeeze()
                update_feat = torch.index_select(search, 0, fore_feat_index).clone() # n, 1, 256
                # update_feat = search.clone()
                self.update_mask_targets(update_feat)

            # print('mask_shape: ', outputs_mask.shape) # 1, 1, 256, 256
            # print('feat raw shape: ', search.shape) # 256, 1, 256
            # assert 0
            

        return outputs_mask, outputs_corner

    def forward_mask(self, Cframe, Cbbox, mode='mask', update_mask_target=False, update_mask_crop=False, update_data=None):
        """
        Args:
            Cframe: Current frame(cv2 array)
            Cbbox: Current bbox (ndarray) (x1,y1,w,h)
        """
        tic = time.time()
        if mode not in ['bbox', 'mask', 'corner', 'all']:
            raise ValueError("mode should be 'bbox' or 'mask' or 'corner' or 'all' ")

        """ Step1: get cropped patch (search region) """
        Cpatch, h_f, w_f = sample_target_SE(Cframe, Cbbox, self.search_factor, self.input_sz,
                                            mode=cv2.BORDER_CONSTANT)
        Cpatch_tensor = self.img_preprocess(Cpatch)

        """ Step2: forward prop (test branch) """
        with torch.no_grad():
            assert self.mask_targets is not None
            output_mask, output_corner = self.generate_test_mask(Cpatch_tensor, update_mask_target)  # (1,1,256,256)  (1, 4)-->[x0, y0, x1, y1]

            # assert 'mask' in output
            Pmask_arr = self.pred2bbox(output_mask, input_type='mask')
            mask = map_mask_back(Cframe, Cbbox, self.search_factor, Pmask_arr, mode=cv2.BORDER_CONSTANT)

            Pbbox_arr = self.pred2bbox(output_corner, input_type='corner')
            # 把bbox的值从256*256 map 回原图尺度
            bbox = self.map_bbox_back(Pbbox_arr, h_f, w_f, Cbbox)
        
        if update_mask_crop:
            if update_data is None:
                assert 0
            else:
                update_frame, update_bbox = update_data
                if update_bbox is None:
                    update_bbox = bbox
            patch1, h_f, w_f = sample_target_SE(update_frame, update_bbox, self.search_factor, self.input_sz,
                                                mode=cv2.BORDER_CONSTANT)
            patch1_tensor = self.img_preprocess(patch1)

            """Step2: get GT's cooridinate on the cropped patch(tensor)"""
            crop_sz = torch.Tensor((self.input_sz, self.input_sz))
            bbox1_tensor = self.gt_preprocess(update_bbox)  # (4,)
            bbox1_crop_tensor = transform_image_to_crop_SE(bbox1_tensor, bbox1_tensor, h_f, w_f, crop_sz).cuda()

            """Step3: forward prop (targets branch)"""
            new_mask_target = self.generate_targets(patch1_tensor, bbox1_crop_tensor)
            self.update_mask_targets(new_mask_target.clone())
            self.online_weight = 0.8 * (1 - self.mask_mem_capacity * 1.0 / 550)


        return mask, bbox

    # w_f = 256 / ws
    def map_bbox_back(self, Pbbox, h_f, w_f, init_bbox, crop_sz=256):
        """
        :param Pbbox: np.array (x0, y0, w, h)
        :param h_f:
        :param w_f:
        :param init_bbox: np.array (x0, y0, w, h)
        :return: Pbbox_init: np.array (x0, y0, w, h)
        """
        Pbbox_init_w = Pbbox[2] / w_f
        Pbbox_init_h = Pbbox[3] / h_f

        init_bbox_center = init_bbox[:2] + 0.5 * init_bbox[2:]
        Pbbox_center = Pbbox[:2] + 0.5 * Pbbox[2:]

        # 在预测图中，预测的中心点距离crop_image中心的距离
        Ptranslation_w = crop_sz * 0.5 - Pbbox_center[0]
        Ptranslation_h = crop_sz * 0.5 - Pbbox_center[1]

        # 根据上面的中心距离映射回原图
        Pbbox_init_center_x = init_bbox_center[0] - Ptranslation_w / w_f
        Pbbox_init_center_y = init_bbox_center[1] - Ptranslation_h / h_f

        Pbbox_init_x0 = Pbbox_init_center_x - 0.5 * Pbbox_init_w
        Pbbox_init_y0 = Pbbox_init_center_y - 0.5 * Pbbox_init_h

        Pbbox_init = np.array([Pbbox_init_x0, Pbbox_init_y0, Pbbox_init_w, Pbbox_init_h])
        return Pbbox_init


    def pred2bbox(self, prediction, input_type=None):
        if input_type == 'mask':
            Pmask = prediction
            Pmask_arr = np.array(Pmask.squeeze().cpu())  # (H,W) (0,1)
            return Pmask_arr
        elif input_type == 'corner':
            Pcorner = prediction  # (x1,y1,x2,y2)
            Pbbox_arr = np.array(Pcorner.squeeze().cpu())
            Pbbox_arr[2:] = Pbbox_arr[2:] - Pbbox_arr[:2]  # (x1,y1,w,h)
            return Pbbox_arr
        else:
            raise ValueError("Input_type should be 'mask' or 'corner'. ")

    def get_network(self, checkpoint_dir):
        network = load_network(checkpoint_dir)
        network.cuda()
        network.eval()
        return network

    def img_preprocess(self, img_arr):
        """ to torch.Tensor(RGB), normalized (minus mean, divided by std)
        Args:
            img_arr: (H,W,3)
        Return:
            (1,1,3,H,W)
        """
        norm_img = ((img_arr / 255.0) - self.mean) / (self.std)
        img_f32 = norm_img.astype(np.float32)
        img_tensor = torch.from_numpy(img_f32).cuda()
        img_tensor = img_tensor.permute((2, 0, 1))
        return img_tensor.unsqueeze(dim=0).unsqueeze(dim=0)

    def gt_preprocess(self, gt_arr):
        """
        Args:
            gt_arr: ndarray (4,)
        Return:
            `torch.Tensor` (4,)
        """
        return torch.from_numpy(gt_arr.astype(np.float32))
